@@ -1,12 +1,13 @@
 #!/usr/bin/perl
 #
-# Copyright © 1996 Andy Guy <awpguy@acs.ucalgary.ca>
+# Copyright © 1996 Andy Guy <andy@cyteen.org>
 # Copyright © 1998 Martin Schulze <joey@infodrom.north.de>
 # Copyright © 1999, 2009 Raphaël Hertzog <hertzog@debian.org>
 #
 # This program is free software; you can redistribute it and/or modify
 # it under the terms of the GNU General Public License as published by
-# the Free Software Foundation; version 2 of the License.
+# the Free Software Foundation; either version 2 of the License, or
+# (at your option) any later version.
 #
 # This program is distributed in the hope that it will be useful,
 # but WITHOUT ANY WARRANTY; without even the implied warranty of
@@ -19,21 +20,22 @@
 use strict;
 use warnings;
 
+use File::Path qw(make_path remove_tree);
+use File::Basename;
+
 eval q{
-    pop @INC if $INC[-1] eq '.';
-    use Net::FTP;
-    use File::Path qw(make_path remove_tree);
-    use File::Basename;
     use File::Find;
     use Data::Dumper;
+
+    use Dpkg::File;
 };
 if ($@) {
-    warn "Please install the 'perl' package if you want to use the\n" .
-         "FTP access method of dselect.\n\n";
+    warn "Missing Dpkg modules required by the FTP access method.\n\n";
     exit 1;
 }
 
-use Dselect::Ftp;
+use Dselect::Method;
+use Dselect::Method::Ftp;
 
 my $ftp;
 
@@ -63,11 +65,7 @@ make_path("$methdir/$CONFIG{dldir}", { mode => 0755 });
 #Read md5sums already calculated
 my %md5sums;
 if (-f "$methdir/md5sums") {
-  local $/;
-  open(my $md5sums_fh, '<', "$methdir/md5sums")
-    or die "couldn't read file $methdir/md5sums";
-  my $code = <$md5sums_fh>;
-  close $md5sums_fh;
+  my $code = file_slurp("$methdir/md5sums");
   my $VAR1; ## no critic (Variables::ProhibitUnusedVariables)
   my $res = eval $code;
   if ($@) {
@@ -76,12 +74,12 @@ if (-f "$methdir/md5sums") {
   if (ref($res)) { %md5sums = %{$res} }
 }
 
-# get a block
+# Get a stanza.
 # returns a ref to a hash containing flds->fld contents
 # white space from the ends of lines is removed and newlines added
 # (no trailing newline).
 # die's if something unexpected happens
-sub getblk {
+sub get_stanza {
     my $fh = shift;
     my %flds;
     my $fld;
@@ -119,7 +117,7 @@ sub procstatus {
     my (%flds, $fld);
     open(my $status_fh, '<', "$vardir/status") or
         die 'Could not open status file';
-    while (%flds = getblk($status_fh), %flds) {
+    while (%flds = get_stanza($status_fh), %flds) {
 	if($flds{'status'} =~ /^install ok/) {
 	    my $cs = (split(/ /, $flds{'status'}))[2];
 	    if (($cs eq 'not-installed') ||
@@ -161,7 +159,7 @@ sub procpkgfile {
     my (@files, @sizes, @md5sums, $pkg, $ver, $nfs, $fld);
     my(%flds);
     open(my $pkgfile_fh, '<', $fn) or die "could not open package file $fn";
-    while (%flds = getblk($pkgfile_fh), %flds) {
+    while (%flds = get_stanza($pkgfile_fh), %flds) {
 	$pkg = $flds{'package'};
 	$ver = $curpkgs{$pkg};
 	@files = split(/[\s\n]+/, $flds{'filename'});
@@ -186,12 +184,12 @@ sub procpkgfile {
 }
 
 print "\nProcessing Package files...\n";
-my ($fn, $i, $j);
+my ($i, $j);
 $i = 0;
 foreach my $site (@{$CONFIG{site}}) {
   $j = 0;
   foreach my $dist (@{$site->[2]}) {
-    $fn = $dist;
+    my $fn = $dist;
     $fn =~ tr#/#_#;
     $fn = "Packages.$site->[0].$fn";
     if (-f $fn) {
@@ -207,7 +205,7 @@ foreach my $site (@{$CONFIG{site}}) {
 
 my $dldir = $CONFIG{dldir};
 # md5sum
-sub md5sum($) {
+sub md5sum {
     my $fn = shift;
     my $m = qx(md5sum $fn);
     $m = (split(' ', $m))[0];
@@ -299,27 +297,23 @@ if($totsize == 0) {
 	    if ($rsize + $totsize > $avsp) {
 		print "no room for: $fn\n";
 		delete $downloads{$fn};
-	    } else {
-		if(yesno($def, $downloads{$fn}
-			 ? "download: $fn ${rsize}k/${csize}k (total = ${totsize}k)"
-			 : "download: $fn ${rsize}k (total = ${totsize}k)")) {
-		    $def = 'y';
-		    $totsize += $rsize;
-		} else {
-		    $def = 'n';
-		    delete $downloads{$fn};
-		}
+            } elsif (yesno($def, $downloads{$fn}
+                           ? "download: $fn ${rsize}k/${csize}k (total = ${totsize}k)"
+                           : "download: $fn ${rsize}k (total = ${totsize}k)")) {
+                $def = 'y';
+                $totsize += $rsize;
+            } else {
+                $def = 'n';
+                delete $downloads{$fn};
 	    }
 	}
     }
 }
 
-sub download() {
-
+sub download {
  my $i = 0;
 
  foreach my $site (@{$CONFIG{site}}) {
-
     my @getfiles = grep { $pkgfiles{$_}[2] == $i } keys %downloads;
     my @pre_dist = (); # Directory to add before $fn
 
@@ -418,6 +412,7 @@ if($totsize != 0) {
               if (yesno('y', "\nDo you want to retry downloading at once")) {
 		  # get the first $fn that foreach would give:
 		  # this is the one that got interrupted.
+                my $fn;
 		MY_ITER: foreach my $ffn (keys(%downloads)) {
 		    $fn = $ffn;
 		    last MY_ITER;
@@ -452,24 +447,24 @@ my %files; # package-version => files...
 # check a deb or split deb file
 # return 1 if it a deb file, 2 if it is a split deb file
 # else 0
-sub chkdeb($) {
+sub chkdeb {
     my ($fn) = @_;
     # check to see if it is a .deb file
-    if(!system("dpkg-deb --info $fn 2>&1 >/dev/null && dpkg-deb --contents $fn 2>&1 >/dev/null")) {
+    if (!system "dpkg-deb --info $fn >/dev/null 2>&1 && dpkg-deb --contents $fn >/dev/null 2>&1") {
 	return 1;
-    } elsif(!system("dpkg-split --info $fn 2>&1 >/dev/null")) {
+    } elsif (!system "dpkg-split --info $fn >/dev/null 2>&1") {
 	return 2;
     }
     return 0;
 }
-sub getdebinfo($) {
+sub getdebinfo {
     my ($fn) = @_;
     my $type = chkdeb($fn);
     my ($pkg, $ver);
     if($type == 1) {
 	open(my $pkgfile_fh, '-|', "dpkg-deb --field $fn")
 	    or die "cannot create pipe for 'dpkg-deb --field $fn'";
-	my %fields = getblk($pkgfile_fh);
+	my %fields = get_stanza($pkgfile_fh);
 	close($pkgfile_fh);
 	$pkg = $fields{'package'};
 	$ver = $fields{'version'};
@@ -489,7 +484,7 @@ sub getdebinfo($) {
 }
 
 # process deb file to make sure we only keep latest versions
-sub prcdeb($$) {
+sub prcdeb {
     my ($dir, $fn) = @_;
     my ($pkg, $ver) = getdebinfo($fn);
     if(!defined($pkg) || !defined($ver)) {
@@ -516,7 +511,7 @@ sub prcdeb($$) {
     }
 }
 
-sub prcfile() {
+sub prcfile {
     my ($fn) = $_;
     if (-f $fn and $fn ne '.') {
         my $dir = '.';
@@ -591,13 +586,11 @@ sub removeinstalled {
 	    my($pkg, $ver) = getdebinfo($fn);
 	    if(!defined($pkg) || !defined($ver)) {
 		print "Could not get info for: $dir/$fn\n";
-	    } else {
-		if ($curpkgs{$pkg} and dcmpvers($ver, 'le', $curpkgs{$pkg})) {
-		    print "deleting: $dir/$fn\n";
-		    unlink $fn;
-		} else {
-		    print "leaving: $dir/$fn\n";
-		}
+            } elsif ($curpkgs{$pkg} and dcmpvers($ver, 'le', $curpkgs{$pkg})) {
+                print "deleting: $dir/$fn\n";
+                unlink $fn;
+            } else {
+                print "leaving: $dir/$fn\n";
 	    }
 	} else {
 	    print "non-debian: $dir/$fn\n";
@@ -624,9 +617,6 @@ foreach my $file (keys %md5sums) {
   next if -f $file;
   delete $md5sums{$file};
 }
-open(my $md5sums_fh, '>', "$methdir/md5sums")
-  or die "can't open $methdir/md5sums in write mode: $!\n";
-print { $md5sums_fh } Dumper(\%md5sums);
-close $md5sums_fh;
+file_dump("$methdir/md5sums", Dumper(\%md5sums));
 
 exit $exit;
